"""Dictionary that automatically adds children dictionaries as necessary
"""

from __future__ import annotations

import json
import os
import pathlib
from typing import Callable


class AutoDict(dict):
  """Dictionary that automatically adds children dictionaries as necessary
  """

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self["__type__"] = "AutoDict"

  def __missing__(self, key: object):
    """Called when a key does not exist in the dictionary

    Args:
      key: Index of item that does not exist

    Returns:
      New AutoDict created at key location
    """
    value = self[key] = AutoDict()
    return value

  def contains(self, *keys: object) -> bool:
    """Check for the presence of keys in dictionary

    Supply multiple keys to check children. Only descends other AutoDicts

    contains("level0", "level1", key) will return True when AutoDict is
    structured as follows:
    {"level0": {"level1": {"key": _ }}}

    Args:
      keys: one or more keys to check for (cascading levels)

    Returns:
      True when key(s) exist
    """
    first_key = keys[0]
    if len(keys) == 1:
      return super().__contains__(first_key)
    if not super().__contains__(first_key):
      return False
    keys = keys[1:]
    if isinstance(self[first_key], AutoDict):
      return self[first_key].contains(*keys)
    if len(keys) == 1:
      return keys[0] in self[first_key]
    return keys in self[first_key]

  def __contains__(self, o: object) -> bool:
    if isinstance(o, list):
      return self.contains(*o)
    return super().__contains__(o)

  @staticmethod
  def decoder(data: dict) -> object:
    """Decode a dictionary object into the appropriate class type

    Args:
      data: dictionary representation of object

    Returns:
      class representation of object
    """
    if "__type__" in data:
      t = data["__type__"]
      if t == "AutoDict":
        return AutoDict(data)
      raise TypeError(f'AutoDict decoder cannot decode __type__="{t}"')
    return data


class JSONAutoDict(AutoDict):
  """AutoDict with json file compatibility/autosaving
  """

  def __init__(self,
               path: str,
               *,
               save_on_exit: bool = False,
               encoder: json.JSONEncoder = None,
               decoder: Callable = AutoDict.decoder,
               **kwargs) -> None:
    """Initialize JSONAutoDict

    Args:
      path: path to json file
      save_on_exit: True will save file when object is closed, False will not
      encoder: JSONEncoder used to serialize the object
      decoder: object_hook used to deserialize the object

      other arguments passed to AutoDict.__init__
    """
    super().__init__(**kwargs)
    self._save_on_exit = save_on_exit
    self._encoder = encoder
    self._decoder = decoder

    self._path = path
    if os.path.exists(path):
      with open(path, "r", encoding="utf-8") as file:
        data = json.load(file, object_hook=self._decoder)
        for k, v in data.items():
          self[k] = v

  def save(self, indent: int = 2) -> None:
    """Write AutoDict to file

    Args:
      indent: Indentation parameter passed to json.dump
    """
    pathlib.Path(self._path).parent.mkdir(parents=True, exist_ok=True)
    with open(self._path, "w", encoding="utf-8") as file:
      json.dump(dict(self), file, cls=self._encoder, indent=indent)

  def __enter__(self) -> AutoDict:
    """Enter ContextManager
    Returns:
      self
    """
    return self

  def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
    """Exit ContextManager
    """
    if self._save_on_exit:
      self.save()
      self._save_on_exit = False

  def __del__(self) -> None:
    """Object destructor
    """
    if self._save_on_exit:
      self.save()



"""Setup and install AutoDict

Typical usage:
  python setup.py develop
  python setup.py install
  python setup.py test
"""

from enum import Flag
import os
from re import MULTILINE
import setuptools
import setuptools.command.build_py
import setuptools.command.develop

module_folder = "autodict"

with open("README.md", encoding="utf-8") as file:
  longDescription = file.read()

required = []

try:
  from tools import gitsemver
  version = gitsemver.get_version()
  with open(f"{module_folder}/version.py", "w", encoding="utf-8") as file:
    file.write('"""Module version information\n"""\n\n')
    file.write(f'version = "{version}"\n')
    file.write(f'version_full = "{version.full_str()}"\n')
except ImportError:
  import re
  with open(f"{module_folder}/version.py", "r", encoding="utf-8") as file:
    version = re.search(r'version = "(.*)"', file.read())[1]

cwd = os.path.dirname(os.path.abspath(__file__))


def find_pyx(path="."):
  pyx_files = []
  for root, _, filenames in os.walk(path):
    for f in filenames:
      if f.endswith(".pyx"):
        pyx_files.append(os.path.join(root, f))
  return pyx_files


def find_cython_extensions(path="."):
  pyx_files = find_pyx(path)
  if len(pyx_files) == 0:
    return []
  import Cython  # pylint: disable=import-outside-toplevel
  extensions = Cython.Build.cythonize(pyx_files, language_level=3)
  if "numpy" in required:
    import numpy  # pylint: disable=import-outside-toplevel

    for ext in extensions:
      ext.include_dirs = [numpy.get_include()]
  return extensions


class BuildPy(setuptools.command.build_py.build_py):

  def run(self):
    setuptools.command.build_py.build_py.run(self)


class Develop(setuptools.command.develop.develop):

  def run(self):
    setuptools.command.develop.develop.run(self)


setuptools.setup(
    name="AutoDict",
    version=str(version),
    description=
    "Dictionary that automatically adds children dictionaries as necessary",
    long_description=longDescription,
    long_description_content_type="text/markdown",
    license="MIT",
    ext_modules=find_cython_extensions(),
    packages=setuptools.find_packages(),
    package_data={module_folder: []},
    install_requires=required,
    extras_require={},
    test_suite="tests",
    scripts=[],
    author="Bradley Davis",
    author_email="me@bradleydavis.tech",
    url="https://github.com/WattsUp/AutoDict",
    download_url=
    "https://github.com/WattsUp/AutoDict/archive/refs/tags/v0.1.0.tar.gz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.7",
    include_package_data=True,
    cmdclass={
        "build_py": BuildPy,
        "develop": Develop,
    },
    zip_safe=False,
)



"""Test module autodict
"""

import os
import pathlib
import random
import string
import unittest

import autodict


def gen_string(min_length: int = 8, max_length: int = 12) -> str:
  """Generate a random string with letter, numbers, and symbols
  Args:
    min_length: minimum length of string
    max_length: maximum length of string
  Returns:
    str Random length string with random characters
  """
  allchar = string.ascii_letters + string.punctuation + string.digits
  return "".join(
      random.choice(allchar)
      for _ in range(random.randint(min_length, max_length)))


class TestAutoDict(unittest.TestCase):
  """Test AutoDict
  """

  _TEST_CONFIG = "testing.json"
  _TEST_ROOT = pathlib.Path(".test")

  def __clean_test_root(self):
    if self._TEST_ROOT.exists():
      for f in os.listdir(self._TEST_ROOT):
        os.remove(self._TEST_ROOT.joinpath(f))
      os.rmdir(self._TEST_ROOT)

  def test_manual_save(self):
    self.__clean_test_root()

    key = gen_string()
    value = gen_string(min_length=40, max_length=50)
    path = str(self._TEST_ROOT.joinpath("autodict.json"))
    with autodict.JSONAutoDict(self._TEST_CONFIG, save_on_exit=True) as c:
      c["manual_save"]["key"] = key
      c["manual_save"]["value"] = value
      c["manual_save"]["path"] = path

    # ContextManager
    with autodict.JSONAutoDict(path, save_on_exit=False) as c:
      c[key] = value
    c = None
    self.assertFalse(os.path.exists(path))

    # No ContextManager
    c = autodict.JSONAutoDict(path, save_on_exit=False)
    c[key] = value
    c = None
    self.assertFalse(os.path.exists(path))

    # Unknown encoding type
    class _UnrecognizedClass:
      pass

    c = autodict.JSONAutoDict(path, save_on_exit=False)
    c[key] = _UnrecognizedClass()
    self.assertRaises(TypeError, c.save)
    c = None
    self.__clean_test_root()

    # Unknown decoding type
    c = autodict.JSONAutoDict(path, save_on_exit=False)
    c[key] = {"__type__": "_UnrecognizedClass"}
    c.save()
    c = None
    self.assertRaises(TypeError,
                      autodict.JSONAutoDict,
                      path,
                      save_on_exit=False)
    self.__clean_test_root()

  def test_auto_save(self):
    self.__clean_test_root()

    section = gen_string()
    key = gen_string()
    value = gen_string(min_length=40, max_length=50)
    path = str(self._TEST_ROOT.joinpath("autodict.json"))
    with autodict.JSONAutoDict(self._TEST_CONFIG, save_on_exit=True) as c:
      c["auto_save"]["section"] = section
      c["auto_save"]["key"] = key
      c["auto_save"]["value"] = value
      c["auto_save"]["path"] = path

    # ContextManager
    with autodict.JSONAutoDict(path, save_on_exit=True) as c:
      c[key] = {key: value}
      c[section][section][key] = value
    self.assertIsInstance(c[key], dict)
    self.assertIsInstance(c[section], autodict.AutoDict)
    c = None
    self.assertTrue(os.path.exists(path))

    with autodict.JSONAutoDict(path, save_on_exit=True) as c:
      self.assertIsInstance(c[key], dict)
      self.assertIsInstance(c[section], autodict.AutoDict)
    c = None
    self.__clean_test_root()

    # No ContextManager
    c = autodict.JSONAutoDict(path, save_on_exit=True)
    c[key] = {key: value}
    c[section][section][key] = value
    c = None
    self.assertTrue(os.path.exists(path))

    self.__clean_test_root()

  def test_contains(self):
    self.__clean_test_root()

    section = gen_string()
    key = gen_string()
    value = gen_string(min_length=40, max_length=50)
    path = str(self._TEST_ROOT.joinpath("autodict.json"))
    with autodict.JSONAutoDict(self._TEST_CONFIG, save_on_exit=True) as c:
      c["contains"]["section"] = section
      c["contains"]["key"] = key
      c["contains"]["value"] = value
      c["contains"]["path"] = path

    # No ContextManager
    c = autodict.JSONAutoDict(path, save_on_exit=True)
    c[key] = {key: value}
    c[section][section][key] = value

    self.assertIn(key, c)
    self.assertNotIn(value, c)
    self.assertTrue(c.contains(section, section, key))
    self.assertFalse(c.contains(section, key, key))
    self.assertIn([key, key], c)
    self.assertNotIn([key, key, key], c)
    c = None

    self.__clean_test_root()
